/**
 * Created by Eike on 15.02.2015.
 */

import bico.BicoPoint;
import bico.CoresetFeature;
import clusterer.KMeansMaster;
import clusterer.KMeansPP;
import measure.Measure;
import org.apache.commons.math3.ml.clustering.CentroidCluster;
import measure.DissimilarityMeasure;
import utils.Pair;
import utils.Triple;
import weka.clusterers.SimpleKMeans;
import weka.core.*;

import java.util.ArrayList;
import java.util.List;

public class Benchmark {
    private SimpleKMeans wekaKmeans;
    private KMeansPP apacheKmeans;
    private int k;
    private int maxIt;

    public Benchmark(int k, int maxIt) throws Exception {
        this.k = k;
        this.maxIt = maxIt;
        wekaKmeans = new SimpleKMeans();
        wekaKmeans.setNumClusters(k);
        wekaKmeans.setMaxIterations(maxIt);
        apacheKmeans = new KMeansPP(k, maxIt);
        Measure.Init(new DissimilarityMeasure());
    }

    public void Bench(ArrayList<BicoPoint> points) throws Exception
    {
        System.out.println("Testing Weka and Apache Kmeans for " + points.size() + " points with {n} centers.");
        FastVector classVal = new FastVector();
        for(int i = 0; i < points.get(0).Dimensions(); ++i)
            classVal.addElement(new Attribute("x" + i));

        Instances instances = new Instances("Test", classVal, 0);
        for(BicoPoint p : points)
        {
            double[] point = new double[p.Dimensions()];
            for(int i = 0; i < point.length; ++i)
                point[i] = p.getEntry(i);
            instances.add(new Instance(1.0, point));
        }
        wekaKmeans.setSeed((int)Math.abs(Math.random() * 10));
        System.out.println("Weka start...");
        long start = System.currentTimeMillis();
        wekaKmeans.buildClusterer(instances);
        Instances result = wekaKmeans.getClusterCentroids();
        long weka = System.currentTimeMillis() - start;
        for(int i = 0; i < result.numInstances(); ++i)
        {
            System.out.println(result.instance(i));
        }

        System.out.println("Weka took: " + weka + "ms to cluster " + points.size() + " points. Purity: ");

        System.out.println("Apache start...");
        start = System.currentTimeMillis();
        List<CentroidCluster<BicoPoint>> res = apacheKmeans.cluster(points);
        long apache = System.currentTimeMillis() - start;

        for(CentroidCluster<BicoPoint> p : res)
        {
            System.out.println(p.getCenter());
        }
        System.out.println("Apache took: " + apache + "ms to cluster " + points.size() + " points. Purity: " + apachePurity(res));

        ArrayList<CoresetFeature> cfs = new ArrayList<CoresetFeature>();
        for(BicoPoint p : points)
            cfs.add(new CoresetFeature(p));
        System.out.println("Bico k-means start...");
        KMeansMaster master = new KMeansMaster(cfs, k, maxIt);
        start = System.currentTimeMillis();
        Pair<Double[][],Triple<Double, Double[], Integer>[]> res2 = master.run();
        for(int i = 0; i < res2.first().length; ++i)
        {
            for(Double d : res2.first()[i])
            {
                System.out.print(d + " ");
            }
            System.out.println();
        }
        System.out.println("Bico k-means took: " + (System.currentTimeMillis() - start) + "ms to cluster " + points.size() + ".");

    }

    public double apachePurity(List<CentroidCluster<BicoPoint>> list)
    {
        double ret = 0.0;

        for(int i = 0; i < list.size(); ++i)
        {
            double max = 0.0;
            for(int j = 0; j < list.size(); ++j)
            {
                if(i==j)
                    continue;
                double tmp = (avgDistanceApache(list.get(i)) + avgDistanceApache(list.get(j))) / norm2(list.get(i).getCenter().getPoint(),list.get(j).getCenter().getPoint());
                if(tmp > max)
                    max = tmp;
            }
            ret += max;
        }

        return ret / list.size();
    }

    public double avgDistanceApache(CentroidCluster<BicoPoint> cluster)
    {
        double ret = 0.0;

        for(int i = 0; i < cluster.getPoints().size(); ++i)
        {
            ret += Math.pow(norm2(cluster.getCenter().getPoint(), cluster.getPoints().get(i).getPoint()), 2.0);
        }

        return Math.sqrt(ret / cluster.getPoints().size());
    }

    public double norm2(double[] point1, double[] point2)
    {
        if(point1.length != point2.length)
            return -1.0;
        double ret = 0.0;
        for(int i = 0; i < point1.length; ++i)
            ret += Math.abs(point1[i] - point2[i]) * Math.abs(point1[i] - point2[i]);
        return Math.sqrt(ret);
    }

    public double wekaPurity(Instances instaces, int size)
    {
        double ret = 0.0;

        for(int i = 0; i < instaces.numInstances(); ++i)
        {
            double max = 0.0;
            for(int j = 0; j < instaces.numInstances(); ++j)
            {
                if(i==j)
                    continue;
                double tmp = (avgDistanceWeka(instaces.instance(i)) + avgDistanceWeka(instaces.instance(j))) / norm2(instaces.instance(i).toDoubleArray(),instaces.instance(j).toDoubleArray());
                if(tmp > max)
                    max = tmp;
            }
            ret += max;
        }

        return ret / (double)size;
    }

    public double avgDistanceWeka(Instance instance)
    {
        double ret = 0.0;
        double[] center = instance.toDoubleArray();
        for(int i = 0; i < instance.dataset().numInstances();++i)
        {
            ret += Math.pow(norm2(center, instance.dataset().instance(i).toDoubleArray()), 2.0);
        }

        return Math.sqrt(ret / instance.numValues());
    }
}
